
/*
 * The Real SoundTracker - module info page
 *
 * Copyright (C) 1998-2019 Michael Krause
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gdk/gdkkeysyms.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <glib/gprintf.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "extspinbutton.h"
#include "gui-subs.h"
#include "gui.h"
#include "instrument-editor.h"
#include "keys.h"
#include "main.h"
#include "module-info.h"
#include "sample-editor.h"
#include "st-subs.h"
#include "track-editor.h"
#include "xm.h"

static GtkWidget *ilist, *slist, *ilist2, *slist2, *songname, *sw1, *sw2;
static GtkWidget *freqmode_w[2], *ptmode_toggle;
static GtkWidget *ins1_label, *smp1_label, *ins2_label, *smp2_label, *mix_label1, *mix_label2;
static guint curi = 0, curs = 0, desti = 0, dests = 0, mix_prev_note;
static gfloat mixlevel = 50.0;

struct tuner {
    GtkWidget *frame, *note_label, *note2_label, *note_spin;
    GtkWidget *period_label, *period_label2, *period_spin, *play, *stop;
    guint note, mode, period;
    gboolean is_ref;
    gint timer;
    const gchar *label_text;
};

static struct tuner *tuning, *ref;

static STSample prev_sample = {{0}};
static guint32 prev_alloc_length = 0;

static gchar* mode1[] = {N_("Single"), N_("Cyclic"), N_("Keyboard")};
static gchar* mode2[] = {N_("Single"), N_("Cyclic"), N_("Coupled")};
enum {
    MODE_SINGLE = 0,
    MODE_CYCLIC,
    MODE_KEYBOARD,
    MODE_COUPLED = MODE_KEYBOARD,
    MODE_LAST
};

static void
ptmode_changed(GtkWidget* widget)
{
    int m = GTK_TOGGLE_BUTTON(widget)->active;
    if (xm) {
        xm->flags &= ~XM_FLAGS_IS_MOD;
        xm->flags |= m * XM_FLAGS_IS_MOD;
    }
    gui_xm_set_modified(1);
}

static void
freqmode_changed(void)
{
    int m = find_current_toggle(freqmode_w, 2);
    if (xm) {
        xm->flags &= ~XM_FLAGS_AMIGA_FREQ;
        xm->flags |= m * XM_FLAGS_AMIGA_FREQ;
    }
    gui_xm_set_modified(1);
}

static void
songname_changed(GtkEntry* entry)
{
    gchar* term;

    g_utf8_strncpy(xm->utf_name, gtk_entry_get_text(entry), 20);
    term = g_utf8_offset_to_pointer(xm->utf_name, 21);
    term[0] = 0;
    xm->needs_conversion = TRUE;
    gui_xm_set_modified(1);
}

static void
modinfo_update_all_samples(const gint n, const gboolean second)
{
    guint i;
    GtkTreeModel* model;
    GtkWidget* list = second ? slist2 : slist;
    GtkListStore* slist_store = GUI_GET_LIST_STORE(list);

    model = gui_list_freeze(list);
    for (i = 0; i < ST_NUM_SAMPLES(&xm->instruments[n]); i++) {
        GtkTreeIter iter;

        if (!gui_list_get_iter(i, slist_store, &iter))
            return; /* Some bullshit happens :-/ */
        gtk_list_store_set(slist_store, &iter, 1,
            xm->instruments[n].samples[i].utf_name, -1);
    }
    gui_list_thaw(list, model);
}

static void
modinfo_update_instrument_full(const gint n, const gboolean second)
{
    GtkTreeIter iter;
    GtkListStore* list_store = GUI_GET_LIST_STORE(second ? ilist2 : ilist);

    if (!gui_list_get_iter(n, list_store, &iter))
        return; /* Some bullshit happens :-/ */
    gtk_list_store_set(list_store, &iter, 1, xm->instruments[n].utf_name,
        2, st_instrument_num_samples(&xm->instruments[n]), -1);

    if (n == (second ? desti : curi))
        modinfo_update_all_samples(n, second);
}

static void
update_labels(struct tuner* t,
    GtkWidget *label,
    GtkWidget *mix_label,
    const gint n,
    const gint ins,
    const gint smp)
{
    gchar *buf;

    buf = g_strdup_printf("%i", n);
    gtk_label_set_text(GTK_LABEL(label), buf);
    g_free(buf);
    buf = g_strdup_printf(_("%sIns. %i, Smp. %i"), t->label_text, ins + 1, smp);
    gtk_frame_set_label(GTK_FRAME(t->frame), buf);
    g_free(buf);
    buf = g_strdup_printf(_("I%i, S%i"), ins + 1, smp);
    gtk_label_set_text(GTK_LABEL(mix_label), buf);
    g_free(buf);
}

static void
ilist_select(GtkTreeSelection* sel)
{
    gint row = gui_list_get_selection_index(sel);

    if (row != -1 && row != curi) {
        curi = row;
        gui_set_current_instrument(row + 1);
        update_labels(tuning, ins1_label, mix_label1, row + 1, curi, curs);
    }
}

static void
slist_select(GtkTreeSelection* sel)
{
    gint row = gui_list_get_selection_index(sel);

    if (row != -1 && row != curs) {
        curs = row;
        gui_set_current_sample(row);
        update_labels(tuning, smp1_label, mix_label1, row, curi, curs);
    }
}

static void
ilist2_select(GtkTreeSelection* sel)
{
    gint row = gui_list_get_selection_index(sel);

    if (row != -1 && row != desti) {
        desti = row;
        modinfo_update_all_samples(row, TRUE);
        update_labels(ref, ins2_label, mix_label2, row + 1, desti, dests);
    }
}

static void
slist2_select(GtkTreeSelection* sel)
{
    gint row = gui_list_get_selection_index(sel);

    if (row != -1 && row != dests) {
        dests = row;
        update_labels(ref, smp2_label, mix_label2, row, desti, dests);
    }
}

static GtkWidget*
add_label(const gchar* title,
    const gchar* value,
    const GtkWidget* box)
{
    GtkWidget *box2, *thing, *frame;

    box2 = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(box), box2, FALSE, FALSE, 0);

    thing = gtk_label_new(_(title));
    gtk_box_pack_start(GTK_BOX(box2), thing, FALSE, FALSE, 0);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_box_pack_end(GTK_BOX(box2), frame, FALSE, FALSE, 0);
    gtk_widget_show(frame);

    thing = gtk_label_new(value);
    gtk_misc_set_alignment(GTK_MISC(thing), 1.0, 0.5);
    gtk_container_add(GTK_CONTAINER(frame), thing);

    return thing;
}

static void
exp_change(GtkExpander *expndr)
{
    gboolean is_on = gtk_expander_get_expanded(expndr);

    (is_on ? gtk_widget_show : gtk_widget_hide)(sw1);
    (is_on ? gtk_widget_show : gtk_widget_hide)(sw2);
}

static void
modinfo_xcopy_ins(const gpointer data)
{
    if (curi != desti) {
        gint action = GPOINTER_TO_INT(data);
        STInstrument* src_ins = &xm->instruments[curi];
        STInstrument* dst_ins = &xm->instruments[desti];

        /* It's simplier just to stop playing instead of performing some doubtful locks */
        gui_play_stop();

        if (action) {
            /* We need some temporary instrument slot to perform the exchange.
               Hopefully in the instrument-editor.c such option already exists.
               The side effect is that the clipboard will be overwritten by the
               second instrument, but I don't regard this as a big harm */
            instrument_editor_copy_instrument(dst_ins);
        }
        st_copy_instrument(src_ins, dst_ins);
        if (action) {
            instrument_editor_paste_instrument(src_ins);
            instrument_editor_update(FALSE);
            sample_editor_update();
            modinfo_update_instrument_full(curi, FALSE);
            modinfo_update_instrument_full(curi, TRUE);
        }
        modinfo_update_instrument_full(desti, FALSE);
        modinfo_update_instrument_full(desti, TRUE);
        gui_xm_set_modified(TRUE);
    }
}

static
void modinfo_update_sample_full(const gint ins, const gint smp, const gboolean second)
{
    GtkTreeIter iter;
    GtkListStore* list_store = GUI_GET_LIST_STORE(second ? slist2 : slist);

    if (!gui_list_get_iter(smp, list_store, &iter))
        return; /* Some bullshit happens :-/ */
    gtk_list_store_set(list_store, &iter, 1,
        xm->instruments[ins].samples[smp].utf_name, -1);
}

static void
modinfo_xcopy_smp(const gpointer data)
{
    if (curi != desti || curs != dests) {
        static STSample* tmp_smp = NULL;

        gint action = GPOINTER_TO_INT(data);
        STInstrument* dest_ins = &xm->instruments[desti];
        STInstrument* src_ins = &xm->instruments[curi];
        STSample* dest_smp = &dest_ins->samples[dests];
        STSample* src_smp = &src_ins->samples[curs];

        if (action) {
            /* We need a place to store one of the samples during the exchange */
            if (!tmp_smp)
                tmp_smp = g_new0(STSample, 1);
            st_copy_sample(dest_smp, tmp_smp);
        }
        if (st_instrument_num_samples(dest_ins) == 0) {
            st_clean_instrument(dest_ins, NULL);
            memset(dest_ins->samplemap, dests, sizeof(dest_ins->samplemap));
        }
        g_mutex_lock(&dest_smp->sample.lock);
        st_copy_sample(src_smp, dest_smp);
        g_mutex_unlock(&dest_smp->sample.lock);
        if (action) {
            if (st_instrument_num_samples(src_ins) == 0) {
                st_clean_instrument(src_ins, NULL);
                memset(src_ins->samplemap, curs, sizeof(src_ins->samplemap));
            }
            g_mutex_lock(&src_smp->sample.lock);
            st_copy_sample(tmp_smp, src_smp);
            g_mutex_unlock(&src_smp->sample.lock);
            sample_editor_update();
            if (curi == desti)
                modinfo_update_sample_full(curi, curs, TRUE);
            modinfo_update_sample_full(curi, curs, FALSE);
        }
        modinfo_update_sample_full(desti, dests, TRUE);
        if (curi == desti)
            modinfo_update_sample_full(desti, dests, FALSE);
        gui_xm_set_modified(TRUE);
    }
}

static void
put_labelled_spin_button(const gchar* title, GtkAdjustment* adj, GtkWidget* box)
{
    GtkWidget *hbox, *thing;

    hbox = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, FALSE, 0);
    thing = gtk_label_new(title);
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);
    thing = extspinbutton_new(adj, 0, 0, TRUE);
    gtk_box_pack_end(GTK_BOX(hbox), thing, FALSE, FALSE, 0);
}

static void
note_changed(GtkSpinButton* spin, struct tuner* t)
{
    t->note = gtk_spin_button_get_value_as_int(spin);
    gtk_label_set_text(GTK_LABEL(t->note_label), instrument_editor_get_note_name(t->note));
}

static void
mix_prev_note_changed(GtkSpinButton* spin, GtkLabel* label)
{
    mix_prev_note = gtk_spin_button_get_value_as_int(spin);
    gtk_label_set_text(label, instrument_editor_get_note_name(mix_prev_note));
}

static void
period_changed(GtkSpinButton* spin, struct tuner* t)
{
    t->period = lrint(gtk_spin_button_get_value(spin) * 1000.0);
}

static void
mode_changed(GtkComboBox* combo, struct tuner* t)
{
    gboolean active;

    t->mode = gtk_combo_box_get_active(combo);

    active = (t->mode == MODE_CYCLIC);
    gtk_widget_set_sensitive(t->period_label, active);
    gtk_widget_set_sensitive(t->period_spin, active);
    gtk_widget_set_sensitive(t->period_label2, active);

    active = !(t->mode == MODE_KEYBOARD); /* == MODE_COUPLED also */
    gtk_widget_set_sensitive(t->play, active);
    gtk_widget_set_sensitive(t->stop, active);
    if (!t->is_ref) {
        gtk_widget_set_sensitive(t->note2_label, active);
        gtk_widget_set_sensitive(t->note_spin, active);
    }
}

static gboolean
play_note(gpointer data)
{
    struct tuner* t = (struct tuner*)data;

    gint ins_num = t->is_ref ? desti : curi;
    gint smp_num = t->is_ref ? dests : curs;
    STSample* sample = &xm->instruments[ins_num].samples[smp_num];

    if (!sample) {
        t->timer = -1;
        return FALSE;
    }

    gui_play_note_full(t->is_ref ? 0 : 1, t->note + 1, sample, 0, sample->sample.length);
    if (!t->is_ref && ref->mode == MODE_COUPLED) {
        sample = &xm->instruments[desti].samples[dests];
        if (sample)
            gui_play_note_full(0, ref->note + 1, sample, 0, sample->sample.length);
    }

    return TRUE;
}

static void
play_clicked(struct tuner* t)
{
    switch (t->mode) {
    case MODE_SINGLE:
        play_note(t);
        break;
    case MODE_CYCLIC:
        if (t->timer == -1) {
            play_note(t);
            t->timer = g_timeout_add(t->period, play_note, t);
        }
        break;
    default:
        break;
    }
}

static void
stop_clicked(struct tuner* t)
{
    if (t->timer != -1) {
        g_source_remove(t->timer);
        t->timer = -1;
    }
    gui_stop_note(t->is_ref ? 0 : 1);
    if (!t->is_ref && ref->mode == MODE_COUPLED)
        gui_stop_note(0);
}

static struct tuner*
tuner_new(const gboolean is_ref, const gchar* label_text)
{
    struct tuner* t = g_new(struct tuner, 1);

    t->note = 48;
    t->mode = 0;
    t->period = 1000;
    t->is_ref = is_ref;
    t->timer = -1;
    t->label_text = label_text;

    t->frame = gtk_frame_new("");
    gtk_frame_set_shadow_type(GTK_FRAME(t->frame), GTK_SHADOW_IN);

    return t;
}

static GtkWidget*
tuner_populate(struct tuner* t)
{
    GtkWidget *vbox, *hbox, *thing;
    GtkListStore* list_store;
    GtkTreeIter iter;
    gint i, width, height;

    vbox = gtk_vbox_new(TRUE, 2);
    hbox = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

    t->play = thing = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(thing), gui_get_pixmap(GUI_PIXMAP_PLAY));
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);
    g_signal_connect_swapped(thing, "clicked", G_CALLBACK(play_clicked), t);

    t->stop = thing = gtk_button_new();
    gtk_container_add(GTK_CONTAINER(thing), gui_get_pixmap(GUI_PIXMAP_STOP));
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);
    g_signal_connect_swapped(thing, "clicked", G_CALLBACK(stop_clicked), t);

    t->period_label = thing = gtk_label_new(_("Period"));
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);
    gtk_widget_set_sensitive(thing, FALSE);
    t->period_label2 = thing = gtk_label_new(_("s"));
    gtk_box_pack_end(GTK_BOX(hbox), thing, FALSE, FALSE, 0);
    gtk_widget_set_sensitive(thing, FALSE);
    t->period_spin = thing = extspinbutton_new(
        GTK_ADJUSTMENT(gtk_adjustment_new(1.0, 0.0, 10.0, 0.1, 1.0, 0.0)),
        0.0, 1, TRUE);
    gtk_box_pack_end(GTK_BOX(hbox), thing, FALSE, FALSE, 0);
    gtk_widget_set_sensitive(thing, FALSE);
    g_signal_connect(thing, "value-changed", G_CALLBACK(period_changed), t);

    hbox = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    thing = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);

    thing = gtk_label_new(_("Mode"));
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);

    list_store = gtk_list_store_new(1, G_TYPE_STRING);
    for (i = 0; i < MODE_LAST; i++) {
        gtk_list_store_append(list_store, &iter);
        gtk_list_store_set(list_store, &iter, 0, _((t->is_ref ? mode2 : mode1)[i]), -1);
    }
    thing = gui_combo_new(list_store);
    gtk_combo_box_set_active(GTK_COMBO_BOX(thing), 0);
    gtk_box_pack_end(GTK_BOX(hbox), thing, FALSE, FALSE, 0);
    g_signal_connect(thing, "changed", G_CALLBACK(mode_changed), t);

    hbox = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
    thing = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);

    t->note2_label = thing = gtk_label_new(_("Note"));
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);

    thing = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(thing), GTK_SHADOW_IN);
    gtk_box_pack_end(GTK_BOX(hbox), thing, FALSE, FALSE, 0);
    t->note_label = gtk_label_new("");
    gui_get_pixel_size(t->note_label, " C#4 ", &width, &height);
    gtk_widget_set_size_request(t->note_label, width, -1);
    gtk_container_add(GTK_CONTAINER(thing), t->note_label);

    t->note_spin = thing = extspinbutton_new(
        GTK_ADJUSTMENT(gtk_adjustment_new(48.0, 0.0, NUM_NOTES - 1, 1.0, 5.0, 0.0)),
        0.0, 0, TRUE);
    gtk_box_pack_end(GTK_BOX(hbox), thing, FALSE, FALSE, 0);
    g_signal_connect(thing, "value-changed", G_CALLBACK(note_changed), t);
    note_changed(GTK_SPIN_BUTTON(thing), t);

    return vbox;
}

static void
mixlevel_changed(GtkAdjustment* adj)
{
    mixlevel = gtk_adjustment_get_value(adj);
}

static void
same_sample_warning(void)
{
    static GtkWidget* dialog = NULL;

    gui_warning_dialog(&dialog, _("Mixing a sample with itself is not a useful idea..."), FALSE);
}

static void
mix_samples(const STSample* dst,
    STSample* sample,
    STSample* sample1,
    const guint32 length1,
    const guint32 length2,
    gfloat mlevel)
{
    guint32 i;
    STSample *rest;

    for (i = 0; i < MIN(length1, length2); i++)
        dst->sample.data[i] = lrintf((float)sample->sample.data[i] * mlevel / 100.0 +
            (float)sample1->sample.data[i] * (100.0 - mlevel) / 100.0);

    if (length1 < length2) {
        mlevel = 100.0 - mlevel;
        rest = sample1;
    } else
        rest = sample;

    for (; i < MAX(length1, length2); i++)
        dst->sample.data[i] = lrintf((float)rest->sample.data[i] * mlevel / 100.0);
}

static void
mix_clicked(void)
{
    STSample* sample = &xm->instruments[curi].samples[curs];
    STSample* sample1 = &xm->instruments[desti].samples[dests];
    guint32 length1 = sample->sample.length, length2 = sample1->sample.length;

    /* Silently ignore empty samples */
    if ((!sample->sample.data) || (!sample1->sample.data))
        return;
    if ((!length1) || (!length2))
        return;

    if (curi == desti && curs == dests) {
        same_sample_warning();
        return;
    }

    g_mutex_lock(&sample->sample.lock);
    if (length1 < length2) {
        sample->sample.data = realloc(sample->sample.data, length2 * sizeof(sample->sample.data[0]));
        sample->sample.length = length2;
    }
    mix_samples(sample, sample, sample1, length1, length2, mixlevel);

    g_mutex_unlock(&sample->sample.lock);
    sample_editor_update();
    gui_xm_set_modified(TRUE);
}

static void
mixer_preview(void)
{
    STSample* sample = &xm->instruments[curi].samples[curs];
    STSample* sample1 = &xm->instruments[desti].samples[dests];
    guint32 length1 = sample->sample.length, length2 = sample1->sample.length, length;
    gpointer sampledata;

    /* Silently ignore empty samples */
    if ((!sample->sample.data) || (!sample1->sample.data))
        return;
    if ((!length1) || (!length2))
        return;

    if (curi == desti && curs == dests) {
        same_sample_warning();
        return;
    }

    length = MAX(length1, length2);
    /* Indeed, we need to update the sample data only if either a mixlevel or one of the samples
       is changed (instrument/sample number or the sample data itself. But tracking the sample
       data modifications (by editing, loading new samples / instrumens / module) is rather
       complicated. So we update te sample data all the time though... */
    if (!prev_sample.sample.data) {
        prev_sample.sample.data = malloc(length * sizeof(sample->sample.data[0]));
        prev_alloc_length = length;
    } else if (prev_alloc_length < length) {
        g_free(prev_sample.sample.data);
        prev_sample.sample.data = malloc(length * sizeof(sample->sample.data[0]));
        prev_alloc_length = length;
    }

    sampledata = prev_sample.sample.data;
    memcpy(&prev_sample, sample, sizeof(prev_sample));
    prev_sample.sample.data = sampledata;
    mix_samples(&prev_sample, sample, sample1, length1, length2, mixlevel);
    prev_sample.sample.length = length;

    gui_play_stop();
    gui_play_note_full(0, mix_prev_note + 1, &prev_sample, 0, prev_sample.sample.length);
}

void modinfo_page_create(GtkNotebook* nb)
{
    GtkWidget *hbox, *thing, *vbox, *expndr, *hbox2, *vbox2, *hbox3, *alignment, *frame, *prev_note_spin;
    GtkAdjustment *adj;
    GtkListStore* list_store;
    GtkTreeIter iter;
    GtkTreeModel* model;
    gchar* ititles[3] = { "n", N_("Instrument Name"), N_("#smpl") };
    gchar* stitles[2] = { "n", N_("Sample Name") };
    GType itypes[3] = { G_TYPE_INT, G_TYPE_STRING, G_TYPE_INT };
    GType stypes[2] = { G_TYPE_INT, G_TYPE_STRING };
    gfloat ialignments[3] = { 0.5, 0.0, 0.5 };
    gfloat salignments[2] = { 0.5, 0.0 };
    gboolean iexpands[3] = { FALSE, TRUE, FALSE };
    gboolean sexpands[3] = { FALSE, TRUE };
    static const gchar* freqlabels[] = { N_("Linear"), N_("Amiga") };
    gint i, width, height;

    vbox = gtk_vbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
    gtk_notebook_append_page(nb, vbox, gtk_label_new(_("Module Info")));
    gtk_widget_show(vbox);

    hbox = gtk_hbox_new(TRUE, 10);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
    gtk_widget_show(hbox);

    /* Main instruments and samples lists */
    ilist = gui_list_in_scrolled_window_full(3, ititles, hbox, itypes, ialignments,
        iexpands, GTK_SELECTION_BROWSE, TRUE, TRUE, &thing);
    gtk_widget_show_all(thing);
    list_store = GUI_GET_LIST_STORE(ilist);
    model = gui_list_freeze(ilist);
    for (i = 1; i <= ST_NUM_INSTRUMENTS(xm); i++) {
        gtk_list_store_append(list_store, &iter);
        gtk_list_store_set(list_store, &iter, 0, i, 1, "", 2, 0, -1);
    }
    gui_list_thaw(ilist, model);
    gui_list_handle_selection(ilist, G_CALLBACK(ilist_select), NULL);

    slist = gui_list_in_scrolled_window_full(2, stitles, hbox, stypes, salignments,
        sexpands, GTK_SELECTION_BROWSE, TRUE, TRUE, &thing);
    gtk_widget_show_all(thing);
    list_store = GUI_GET_LIST_STORE(slist);
    model = gui_list_freeze(slist);
    for (i = 0; i < ST_NUM_SAMPLES(&xm->instruments[0]); i++) {
        gtk_list_store_append(list_store, &iter);
        gtk_list_store_set(list_store, &iter, 0, i, 1, "", -1);
    }
    gui_list_thaw(slist, model);
    gui_list_handle_selection(slist, G_CALLBACK(slist_select), NULL);

    /* Secondary instruments and samples lists for extended functions */
    ilist2 = gui_list_in_scrolled_window_full(3, ititles, hbox, itypes, ialignments,
        iexpands, GTK_SELECTION_BROWSE, TRUE, TRUE, &sw1);
    gtk_widget_show(ilist2);
    list_store = GUI_GET_LIST_STORE(ilist2);
    model = gui_list_freeze(ilist2);
    for (i = 1; i <= ST_NUM_INSTRUMENTS(xm); i++) {
        gtk_list_store_append(list_store, &iter);
        gtk_list_store_set(list_store, &iter, 0, i, 1, "", 2, 0, -1);
    }
    gui_list_thaw(ilist2, model);
    gui_list_handle_selection(ilist2, G_CALLBACK(ilist2_select), NULL);

    slist2 = gui_list_in_scrolled_window_full(2, stitles, hbox, stypes, salignments,
        sexpands, GTK_SELECTION_BROWSE, TRUE, TRUE, &sw2);
    gtk_widget_show(slist2);
    list_store = GUI_GET_LIST_STORE(slist2);
    model = gui_list_freeze(slist2);
    for (i = 0; i < ST_NUM_SAMPLES(&xm->instruments[0]); i++) {
        gtk_list_store_append(list_store, &iter);
        gtk_list_store_set(list_store, &iter, 0, i, 1, "", -1);
    }
    gui_list_thaw(slist2, model);
    gui_list_handle_selection(slist2, G_CALLBACK(slist2_select), NULL);

    /* Module metadata */
    hbox = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

    thing = gtk_label_new(_("Songname:"));
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, TRUE, 0);

    gui_get_text_entry(20, songname_changed, &songname);
    gtk_box_pack_start(GTK_BOX(hbox), songname, TRUE, TRUE, 0);

    add_empty_hbox(hbox);
    thing = make_labelled_radio_group_box(_("Frequencies:"), freqlabels, freqmode_w, freqmode_changed);
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, TRUE, 0);
    add_empty_hbox(hbox);

    ptmode_toggle = gtk_check_button_new_with_label(_("ProTracker Mode"));
    gtk_box_pack_start(GTK_BOX(hbox), ptmode_toggle, FALSE, TRUE, 0);
    gtk_widget_show(ptmode_toggle);
    g_signal_connect(ptmode_toggle, "toggled",
        G_CALLBACK(ptmode_changed), NULL);

    add_empty_hbox(hbox);
    gtk_widget_show_all(hbox);

    thing = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(vbox), thing, FALSE, FALSE, 0);
    gtk_widget_show(thing);

    /* Extended editor */
    expndr = gtk_expander_new(_("Extended Instrument/Sample Editor"));
    gtk_expander_set_expanded(GTK_EXPANDER(expndr), FALSE);
    gtk_box_pack_end(GTK_BOX(vbox), expndr, FALSE, FALSE, 0);
    g_signal_connect(expndr, "notify::expanded", G_CALLBACK(exp_change), NULL);
    hbox = gtk_hbox_new(FALSE, 4);
    gtk_container_add(GTK_CONTAINER(expndr), hbox);

    vbox = gtk_vbox_new(TRUE, 2);
    gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
    ins1_label = add_label(N_("Ins. 1"), "1", vbox);
    gui_get_pixel_size(ins1_label, "000", &width, &height);
    gtk_widget_set_size_request(ins1_label, width, -1);
    smp1_label = add_label(N_("Smp. 1"), "0", vbox);
    gtk_widget_set_size_request(smp1_label, width, -1);
    ins2_label = add_label(N_("Ins. 2"), "1", vbox);
    gtk_widget_set_size_request(ins2_label, width, -1);
    smp2_label = add_label(N_("Smp. 2"), "0", vbox);
    gtk_widget_set_size_request(smp2_label, width, -1);

    thing = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);

    vbox = gtk_vbox_new(TRUE, 2);
    gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);

    thing = gtk_button_new_with_label(_("I1 => I2"));
    gtk_widget_set_tooltip_text(thing, _("Copy Instrument 1 to Instrument 2"));
    gtk_box_pack_start(GTK_BOX(vbox), thing, FALSE, FALSE, 0);
    g_signal_connect_swapped(thing, "clicked", G_CALLBACK(modinfo_xcopy_ins), GINT_TO_POINTER(0));

    thing = gtk_button_new_with_label(_("I1 <=> I2"));
    gtk_widget_set_tooltip_text(thing, _("Exchange Instruments 1 and 2"));
    gtk_box_pack_start(GTK_BOX(vbox), thing, FALSE, FALSE, 0);
    g_signal_connect_swapped(thing, "clicked", G_CALLBACK(modinfo_xcopy_ins), GINT_TO_POINTER(1));

    thing = gtk_button_new_with_label(_("S1 => S2"));
    gtk_widget_set_tooltip_text(thing, _("Copy Sample 1 to Sample 2"));
    gtk_box_pack_start(GTK_BOX(vbox), thing, FALSE, FALSE, 0);
    g_signal_connect_swapped(thing, "clicked", G_CALLBACK(modinfo_xcopy_smp), GINT_TO_POINTER(0));

    thing = gtk_button_new_with_label(_("S1 <=> S2"));
    gtk_widget_set_tooltip_text(thing, _("Exchange Samples 1 and 2"));
    gtk_box_pack_start(GTK_BOX(vbox), thing, FALSE, FALSE, 0);
    g_signal_connect_swapped(thing, "clicked", G_CALLBACK(modinfo_xcopy_smp), GINT_TO_POINTER(1));

    thing = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);

    /* Mixer */

    vbox = gtk_vbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 0);
    thing = gtk_label_new(_("Mixing balance"));
    gtk_misc_set_alignment(GTK_MISC(thing), 0.5, 0.5);
    gtk_box_pack_start(GTK_BOX(vbox), thing, FALSE, TRUE, 0);

    hbox2 = gtk_hbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(vbox), hbox2, FALSE, TRUE, 0);
    vbox2 = gtk_vbox_new(FALSE, 2);
    gtk_box_pack_start(GTK_BOX(hbox2), vbox2, TRUE, TRUE, 0);
    hbox3 = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox2), hbox3, FALSE, TRUE, 0);

    mix_label1 = gtk_label_new(_("I1, S0"));
    gtk_box_pack_start(GTK_BOX(hbox3), mix_label1, FALSE, TRUE, 0);

    mix_label2 = gtk_label_new(_("I1, S0"));
    gtk_box_pack_end(GTK_BOX(hbox3), mix_label2, FALSE, TRUE, 0);

    adj = GTK_ADJUSTMENT(gtk_adjustment_new(50.0, 0.0, 100.0, 1.0, 10.0, 0.0));
    thing = gtk_hscale_new(adj);
    gtk_scale_set_draw_value(GTK_SCALE(thing), FALSE);
    gtk_box_pack_start(GTK_BOX(vbox2), thing, FALSE, TRUE, 0);
    g_signal_connect(adj, "value-changed", G_CALLBACK(mixlevel_changed), NULL);

    alignment = gtk_alignment_new(0.5, 1.0, 0.0, 0.0);
    gtk_box_pack_start(GTK_BOX(hbox2), alignment, FALSE, FALSE, 0);
    hbox3 = gtk_hbox_new(FALSE, 2);
    gtk_container_add(GTK_CONTAINER(alignment), hbox3);

    thing = extspinbutton_new(adj, 0.0, 0, TRUE);
    gtk_box_pack_start(GTK_BOX(hbox3), thing, FALSE, FALSE, 0);
    thing = gtk_label_new("%");
    gtk_box_pack_start(GTK_BOX(hbox3), thing, FALSE, FALSE, 0);

    hbox2 = gtk_hbox_new(FALSE, 4);
    gtk_box_pack_end(GTK_BOX(vbox), hbox2, FALSE, TRUE, 0);

    thing = gtk_label_new(_("Note"));
    gtk_widget_set_tooltip_text(thing, _("Note to play preview"));
    gtk_box_pack_start(GTK_BOX(hbox2), thing, FALSE, FALSE, 0);

    prev_note_spin = extspinbutton_new(
        GTK_ADJUSTMENT(gtk_adjustment_new(48.0, 0.0, NUM_NOTES - 1, 1.0, 5.0, 0.0)),
        0.0, 0, TRUE);
    gtk_box_pack_start(GTK_BOX(hbox2), prev_note_spin, FALSE, FALSE, 0);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_box_pack_start(GTK_BOX(hbox2), frame, FALSE, FALSE, 0);
    thing = gtk_label_new("");
    gui_get_pixel_size(thing, " C#4 ", &width, &height);
    gtk_widget_set_size_request(thing, width, -1);
    gtk_container_add(GTK_CONTAINER(frame), thing);
    g_signal_connect(prev_note_spin, "value-changed", G_CALLBACK(mix_prev_note_changed), thing);
    mix_prev_note_changed(GTK_SPIN_BUTTON(prev_note_spin), GTK_LABEL(thing));

    thing = gtk_button_new_with_label(_("Preview"));
    gtk_widget_set_tooltip_text(thing, _("Play the mixed sample without module modification"));
    gtk_box_pack_start(GTK_BOX(hbox2), thing, TRUE, TRUE, 0);
    g_signal_connect(thing, "clicked", G_CALLBACK(mixer_preview), NULL);

    thing = gtk_button_new_with_label(_("Mix!"));
    gtk_widget_set_tooltip_text(thing,
        _("Mix sample 1 of instrument 1 with sample 2 of instrument 2 with the given balance ratio"));
    gtk_box_pack_start(GTK_BOX(hbox2), thing, TRUE, TRUE, 0);
    g_signal_connect(thing, "clicked", G_CALLBACK(mix_clicked), NULL);

    thing = gtk_vseparator_new();
    gtk_box_pack_start(GTK_BOX(hbox), thing, FALSE, FALSE, 0);

    /* Tuning facility, tuning sample */
    tuning = tuner_new(FALSE, _("Tuning: "));
    gtk_box_pack_start(GTK_BOX(hbox), tuning->frame, FALSE, FALSE, 0);
    update_labels(tuning, ins1_label, mix_label1, 1, 0, 0);

    hbox2 = gtk_hbox_new(FALSE, 4);
    gtk_container_set_border_width(GTK_CONTAINER(hbox2), 4);
    gtk_container_add(GTK_CONTAINER(tuning->frame), hbox2);
    vbox = gtk_vbox_new(TRUE, 2);
    gtk_box_pack_start(GTK_BOX(hbox2), vbox, FALSE, FALSE, 0);

    thing = gtk_hscale_new(sample_editor_get_adjustment(SAMPLE_EDITOR_FINETUNE));
    gtk_scale_set_draw_value(GTK_SCALE(thing), FALSE);
    gtk_box_pack_start(GTK_BOX(vbox), thing, FALSE, FALSE, 0);
    put_labelled_spin_button(_("Finetune"), sample_editor_get_adjustment(SAMPLE_EDITOR_FINETUNE), vbox);
    put_labelled_spin_button(_("RelNote"), sample_editor_get_adjustment(SAMPLE_EDITOR_RELNOTE), vbox);

    vbox = tuner_populate(tuning);
    gtk_box_pack_start(GTK_BOX(hbox2), vbox, FALSE, FALSE, 0);

    /* Reference sample */
    ref = tuner_new(TRUE, _("Reference: "));
    gtk_box_pack_start(GTK_BOX(hbox), ref->frame, FALSE, FALSE, 0);

    vbox = tuner_populate(ref);
    gtk_container_set_border_width(GTK_CONTAINER(vbox), 4);
    gtk_container_add(GTK_CONTAINER(ref->frame), vbox);
    update_labels(ref, ins2_label, mix_label2, 1, 0, 0);

    gtk_widget_show_all(expndr);
}

void modinfo_delete_unused_instruments(void)
{
    int i;

    for (i = 0; i < ST_NUM_INSTRUMENTS(xm); i++) {
        if (!st_instrument_used_in_song(xm, i + 1)) {
            st_clean_instrument(&xm->instruments[i], NULL);
            gui_xm_set_modified(1);
        }
    }

    sample_editor_update();
    instrument_editor_update(FALSE);
    modinfo_update_all();
}

gboolean
modinfo_page_handle_keys(int shift,
    int ctrl,
    int alt,
    guint32 keyval,
    gboolean pressed)
{
    gint modifiers = ENCODE_MODIFIERS(shift, ctrl, alt);
    gint i = keys_get_key_meaning(keyval, modifiers, -1);
    gboolean handled = FALSE, samples_focused;

    if (i != -1 && KEYS_MEANING_TYPE(i) == KEYS_MEANING_NOTE) {
        if (tuning->mode == MODE_KEYBOARD) {
            STSample* sample = &xm->instruments[curi].samples[curs];

            handled = gui_play_note_no_repeat(keyval, modifiers, pressed,
                sample, 0, sample->sample.length, 1, FALSE, NULL);

            /* handled means that it's a real, not fake, keypress and the note is correct */
            if (ref->mode == MODE_COUPLED && handled) {
                if (pressed) {
                    sample = &xm->instruments[desti].samples[dests];
                    if (sample)
                        gui_play_note_full(0, ref->note + 1, sample, 0, sample->sample.length);
                } else
                    gui_stop_note(0);
            }
        } else
            track_editor_do_the_note_key(i, pressed, keyval, ENCODE_MODIFIERS(shift, ctrl, alt), TRUE);
        return TRUE;
    }

    if (!pressed)
        return FALSE;

    samples_focused = (GTK_WINDOW(mainwindow)->focus_widget == slist);
    switch (keyval) {
    case GDK_Tab:
    case GDK_ISO_Left_Tab:
        gtk_window_set_focus(GTK_WINDOW(mainwindow), samples_focused ? ilist : slist);
        handled = TRUE;
        break;
    case GDK_Up:
        if (samples_focused)
            gui_offset_current_sample(shift ? -4 : -1);
        else
            gui_offset_current_instrument(shift ? -5 : -1);
        handled = TRUE;
        break;
    case GDK_Down:
        if (samples_focused)
            gui_offset_current_sample(shift ? 4 : 1);
        else
            gui_offset_current_instrument(shift ? 5 : 1);
        handled = TRUE;
        break;
    case GDK_Delete:
        handled = instrument_editor_clear_current_instrument();
        break;
    default:
        break;
    }
    return handled;
}

void modinfo_stop_cycling(void)
{
    if (tuning->timer != -1) {
        g_source_remove(tuning->timer);
        tuning->timer = -1;
    }
    if (ref->timer != -1) {
        g_source_remove(ref->timer);
        ref->timer = -1;
    }
}

void
modinfo_update_instrument(int n)
{
    modinfo_update_instrument_full(n, FALSE);
}

void modinfo_update_sample(int n)
{
    modinfo_update_sample_full(curi, n, FALSE);
}

void modinfo_update_all(void)
{
    int i;
    int m = xm_get_modified();

    for (i = 0; i < ST_NUM_INSTRUMENTS(xm); i++) {
        modinfo_update_instrument_full(i, FALSE);
        modinfo_update_instrument_full(i, TRUE);
    }

    gtk_entry_set_text(GTK_ENTRY(songname), xm->utf_name);

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ptmode_toggle), xm->flags & XM_FLAGS_IS_MOD);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(freqmode_w[xm->flags & XM_FLAGS_AMIGA_FREQ]), TRUE);
    gui_xm_set_modified(m);
}

void modinfo_set_current_instrument(int n)
{
    g_return_if_fail(n >= 0 && n < ST_NUM_INSTRUMENTS(xm));

    gui_list_select(ilist, n, FALSE, 0.0);
    modinfo_update_all_samples(n, FALSE);
}

void modinfo_set_current_sample(int n)
{
    g_return_if_fail(n >= 0 && n < ST_NUM_SAMPLES(&xm->instruments[0]));

    gui_list_select(slist, n, FALSE, 0.0);
}

gint modinfo_get_current_sample(void)
{
    return curs;
}

void modinfo_find_unused_pattern(void)
{
    int n = st_find_first_unused_and_empty_pattern(xm);

    if (n != -1)
        gui_set_current_pattern(n, TRUE);
}

void modinfo_copy_to_unused_pattern(void)
{
    int n = st_find_first_unused_and_empty_pattern(xm);
    int c = gui_get_current_pattern();

    if (n != -1 && !st_is_empty_pattern(&xm->patterns[c])) {
        gui_play_stop();
        st_copy_pattern(&xm->patterns[n], &xm->patterns[c]);
        gui_xm_set_modified(1);
        gui_set_current_pattern(n, TRUE);
    }
}

// Move unused patterns to the end of the pattern space
void modinfo_pack_patterns(void)
{
    int i, j, last, used;

    // Get number of last used pattern and number of used patterns
    for (i = 0, last = 0, used = 0; i < ST_NUM_PATTERNS(xm); i++) {
        if (st_is_pattern_used_in_song(xm, i)) {
            last = i;
            used++;
        }
    }

    // Put unused patterns to the end
    for (i = 0; i < used;) {
        if (!st_is_pattern_used_in_song(xm, i)) {
            for (j = i; j < last; j++)
                st_exchange_patterns(xm, j, j + 1);
        } else {
            i++;
        }
    }

    gui_playlist_initialize();
    gui_xm_set_modified(1);
}

// Put patterns in playback order, move unused patterns to the end of the pattern space
void modinfo_reorder_patterns(void)
{
    modinfo_pack_patterns();
    gui_xm_set_modified(1);
}

// Clear patterns which are not in the playlist
void modinfo_clear_unused_patterns(void)
{
    int i;

    for (i = 0; i < ST_NUM_PATTERNS(xm); i++)
        if (!st_is_pattern_used_in_song(xm, i))
            st_clear_pattern(&xm->patterns[i]);

    tracker_redraw(tracker);
    gui_xm_set_modified(1);
}

// Optimize -- clear everything unused

void modinfo_optimize_module(void)
{
    gchar* infbuf;
    int a, b, c, d, e;

    d = ST_NUM_PATTERNS(xm);
    e = ST_NUM_INSTRUMENTS(xm);
    for (a = 0, b = 0, c = 0; a < d; a++) {
        if (!st_is_pattern_used_in_song(xm, a))
            b++;

        if (a < e)
            if (!st_instrument_used_in_song(xm, a + 1))
                c++;
    }

    infbuf = g_strdup_printf(_("Unused patterns: %d (used: %d)\nUnused instruments: %d (used: %d)\n\nClear unused and reorder playlist?\n"),
        b, d - b, c, e - c);

    if (gui_ok_cancel_modal(mainwindow, infbuf)) {
        modinfo_clear_unused_patterns();
        modinfo_delete_unused_instruments();
        modinfo_reorder_patterns();
        gui_xm_set_modified(1);
    }
    g_free(infbuf);
}
