
/*
 * The Real SoundTracker - main window oscilloscope group
 *
 * Copyright (C) 1998-2019 Michael Krause
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <glib/gprintf.h>

#include "scalablepic.h"
#include "audio.h"
#include "colors.h"
#include "gui-settings.h"
#include "gui-subs.h"
#include "sample-display.h"
#include "scope-group.h"
#include "draw-interlayer.h"

struct _NumberObject{
    DIGC mask_gc, mask_bg_gc;
    DIDrawable drawable;
    GdkPixmap *number_pm;
    /* Size of the drawing area */
    gint width, height;
    /* Size of the pixmap */
    gint pm_width, pm_height;
    gint number;
    gboolean firsttime;
    unsigned int old_pixel;
};

G_DEFINE_TYPE(ScopeGroup, scope_group, GTK_TYPE_HBOX)

/* Displacement between the channel number and the left top corner:
   x_displ = XDISP + width / FRACTION, the same is for y coordinate */
static guint XDISP = 1, YDISP = 1, FRACTION = 200;

static void
scope_group_set_channel_state(ScopeGroup *s,
    const guint n,
    const gboolean on)
{
    if (s->scopes_on) {
        gtk_widget_hide(on ? s->mutedpic[n] : GTK_WIDGET(s->scopes[n]));
        gtk_widget_show(on ? GTK_WIDGET(s->scopes[n]) : s->mutedpic[n]);
    } else {
        gtk_widget_hide(on ? s->mutedpic[n] : s->unmutedpic[n]);
        gtk_widget_show(on ? s->unmutedpic[n] : s->mutedpic[n]);
    }
}

void
scope_group_toggle_channel(ScopeGroup *s,
    const guint n)
{
    const guint mask = 1 << n;
    gboolean on = ((s->on_mask & mask) != 0);

    player_mute_channels[n] = on;
    s->on_mask ^= mask;

    scope_group_set_channel_state(s, n, !on);
}

void
scope_group_solo_channel(ScopeGroup *sg,
    const guint n)
{
    gint i;

    for (i = 0; i < sg->numchan; i++) {
        gboolean on = ((sg->on_mask & (1 << i)) != 0);

        if ((i == n) != on)
            scope_group_set_channel_state(sg, i, i == n);
        player_mute_channels[i] = (i != n);
    }

    sg->on_mask = 1 << n;
}

void
scope_group_all_channels_on(ScopeGroup *sg)
{
    gint i;

    for (i = 0; i < sg->numchan; i++)
        scope_group_set_channel_state(sg, i, TRUE);
    sg->on_mask = 0xFFFFFFFF;
    memset(player_mute_channels, 0, sizeof(player_mute_channels));
}

static gboolean
scope_group_button_press(GtkWidget* widget,
    GdkEventButton* event,
    gpointer data)
{
    GtkWidget* w;
    ScopeGroup* sg;
    const gint n = GPOINTER_TO_INT(data);
    static guint prev_time = 0;

    /* A strange Gtk+-2 bug, when one real event produces 2 callbacks */
    if (event->time == prev_time)
        return FALSE;
    prev_time = event->time;

    /* enjoy the hack!!! */
    g_return_val_if_fail((w = widget->parent) != NULL, FALSE);
    sg = SCOPE_GROUP(w->parent); /* button<-table<-scope_group */

    if (event->button == 1) {
        scope_group_toggle_channel(sg, n);
        return TRUE;
    } else if (event->button == 3) {
        /* Turn all scopes on if some were muted, or make solo selected otherwise */
        if (sg->on_mask == 0xFFFFFFFF)
            scope_group_solo_channel(sg, n);
        else
            scope_group_all_channels_on(sg);
        return TRUE;
    }

    return FALSE;
}

void scope_group_set_num_channels(ScopeGroup* s,
    int num_channels)
{
    int i;

    // Remove superfluous scopes from table
    for (i = num_channels; i < s->numchan; i++) {
        gtk_container_remove(GTK_CONTAINER(s->table), s->scopebuttons[i]);
    }

    // Resize table
    gtk_table_resize(GTK_TABLE(s->table), 2, num_channels / 2);

    // Add new scopes to table
    for (i = s->numchan; i < num_channels; i++) {
        g_object_ref(G_OBJECT(s->scopebuttons[i]));
        gtk_table_attach_defaults(GTK_TABLE(s->table), s->scopebuttons[i], i / 2, i / 2 + 1, i % 2, i % 2 + 1);
    }

    s->numchan = num_channels;
    s->on_mask = 0xFFFFFFFF; /* all channels are on */
    memset(player_mute_channels, 0, sizeof(player_mute_channels));

    // Reset all buttons (enable all channels)
    for (i = 0; i < 32; i++)
        scope_group_set_channel_state(s, i, TRUE);
}

void scope_group_enable_scopes(ScopeGroup* s,
    int enable)
{
    int i;

    s->scopes_on = enable;

    for (i = 0; i < 32; i++) {
        if (s->on_mask & (1 << i)) {
            gtk_widget_hide(enable ? s->unmutedpic[i] : GTK_WIDGET(s->scopes[i]));
            gtk_widget_show(enable ? GTK_WIDGET(s->scopes[i]) : s->unmutedpic[i]);
        }
    }
}

static gint
scope_group_timeout(ScopeGroup* s)
{
    double time1, time2;
    int i, l;
    static gint16* buf = NULL;
    static int bufsize = 0;
    int o1, o2;

    if (!s->scopes_on || !scopebuf_ready || !current_driver)
        return TRUE;

    time1 = current_driver->get_play_time(current_driver_object);
    time2 = time1 + (double)1 / s->update_freq;

    for (i = 0; i < 2; i++) {
        if (time1 < scopebuf_start.time || time2 < scopebuf_start.time) {
            // requesting too old samples -- scopebuf_length is too low
            goto ende;
        }

        if (time1 >= scopebuf_end.time || time2 >= scopebuf_end.time) {
            /* requesting samples which haven't been even rendered yet.
	       can happen with short driver latencies. */
            time1 -= (double)1 / s->update_freq;
            time2 -= (double)1 / s->update_freq;
        } else {
            break;
        }
    }

    if (i == 2) {
        goto ende;
    }

    o1 = (time1 - scopebuf_start.time) * scopebuf_freq + scopebuf_start.offset;
    o2 = (time2 - scopebuf_start.time) * scopebuf_freq + scopebuf_start.offset;

    l = o2 - o1;
    if (bufsize < l) {
        free(buf);
        buf = malloc(2 * l);
        bufsize = l;
    }

    o1 %= scopebuf_length;
    o2 %= scopebuf_length;
    g_assert(o1 >= 0 && o1 <= scopebuf_length);
    g_assert(o2 >= 0 && o2 <= scopebuf_length);

    for (i = 0; i < s->numchan; i++) {
        if (o2 > o1) {
            sample_display_set_data(s->scopes[i], scopebufs[i] + o1, ST_MIXER_FORMAT_S16_LE, l, TRUE);
        } else {
            memcpy(buf, scopebufs[i] + o1, 2 * (scopebuf_length - o1));
            memcpy(buf + scopebuf_length - o1, scopebufs[i], 2 * o2);
            sample_display_set_data(s->scopes[i], buf, ST_MIXER_FORMAT_S16_LE, l, TRUE);
        }
    }

    return TRUE;

ende:
    for (i = 0; i < s->numchan; i++) {
        sample_display_set_data(s->scopes[i], NULL, ST_MIXER_FORMAT_S8, 0, FALSE);
    }
    return TRUE;
}

void scope_group_start_updating(ScopeGroup* s)
{
    if (!s->scopes_on || s->gtktimer != -1)
        return;

    s->gtktimer = g_timeout_add(1000 / s->update_freq, (gint(*)())scope_group_timeout, s);
}

void scope_group_stop_updating(ScopeGroup* s)
{
    int i;

    if (!s->scopes_on || s->gtktimer == -1)
        return;

    g_source_remove(s->gtktimer);
    s->gtktimer = -1;

    for (i = 0; i < s->numchan; i++) {
        sample_display_set_data(s->scopes[i], NULL, ST_MIXER_FORMAT_S8, 0, FALSE);
    }
}

void scope_group_set_update_freq(ScopeGroup* s,
    int freq)
{
    s->update_freq = freq;
    if (s->scopes_on && s->gtktimer != -1) {
        scope_group_stop_updating(s);
        scope_group_start_updating(s);
    }
}

static void
number_sub_draw(cairo_t* cr, PangoLayout* layout, gdouble x, gdouble y)
{
    cairo_move_to(cr, x, y);
    pango_cairo_update_layout(cr, layout);
    pango_cairo_show_layout(cr, layout);
}

/* Making a pixmap cache for each channel number */
static void
number_style_set(GtkWidget* widget, gpointer data)
{
    if (gtk_widget_get_realized(widget)) {
        PangoContext* context;
        cairo_t* cr;
        cairo_font_options_t* cfo;
        PangoLayout* layout;
        GdkColor* num_color;
        gchar num_str[4]; /* I hope we don't need more that 3 digits */

        NumberObject* no=(NumberObject*)data;

        context = gtk_widget_create_pango_context(widget);
        layout = pango_layout_new(context);
        g_object_unref(context);
        g_snprintf(num_str, 3, "%02d", no->number);
        pango_layout_set_font_description(layout, gtk_widget_get_style(widget)->font_desc);
        pango_layout_set_text(layout, num_str, -1);
        pango_layout_get_pixel_size(layout, &no->pm_width, &no->pm_height);
        no->pm_width += 2; /* For outline */
        no->pm_height += 2;

        if (no->number_pm)
            g_object_unref(no->number_pm);
        no->number_pm = gdk_pixmap_new(widget->window, no->pm_width, no->pm_height, -1);
        no->drawable = di_get_drawable(no->number_pm);
        cr = gdk_cairo_create(no->number_pm);

        /* Switching off antialiasing improves channel numbers look making them brighter */
        cfo = cairo_font_options_create();
        cairo_get_font_options(cr, cfo);
        cairo_font_options_set_antialias(cfo, CAIRO_ANTIALIAS_NONE);
        cairo_set_font_options(cr, cfo);
        cairo_font_options_destroy(cfo);

        di_draw_rectangle(no->drawable, no->mask_bg_gc,
            TRUE, 0, 0, no->pm_width, no->pm_height);
        num_color = colors_get_color(COLOR_CHANNUMS);
        no->old_pixel = num_color->pixel;

        /* Outline */
        gdk_cairo_set_source_color(cr, colors_get_color(COLOR_UNMUTED_BG));
        pango_cairo_update_layout(cr, layout);
        pango_cairo_show_layout(cr, layout);
        number_sub_draw(cr, layout, 1.0, 0.0);
        number_sub_draw(cr, layout, 2.0, 0.0);
        number_sub_draw(cr, layout, 2.0, 1.0);
        number_sub_draw(cr, layout, 2.0, 2.0);
        number_sub_draw(cr, layout, 1.0, 2.0);
        number_sub_draw(cr, layout, 0.0, 2.0);
        number_sub_draw(cr, layout, 0.0, 1.0);

        /* Number */
        gdk_cairo_set_source_color(cr, num_color);
        number_sub_draw(cr, layout, 1.0, 1.0);

        cairo_destroy(cr);
        g_object_unref(layout);
    }
}

static NumberObject*
init_number(const gint number, GtkWidget* widget)
{
    NumberObject *no = g_new(NumberObject, 1);

    no->width = -1;
    no->height = -1;
    no->number = number;
    no->number_pm = NULL;
    no->firsttime = TRUE;
    no->old_pixel = 0;

    return no;
}

static void
number_realize(GtkWidget* widget, gpointer data)
{
    NumberObject *no = (NumberObject*)data;

    if (no->firsttime) {
        no->firsttime = FALSE;
        no->mask_gc = di_gc_new(widget->window);
        no->mask_bg_gc = colors_get_gc(COLOR_BLACK);
    }

    number_style_set(widget, data);
}

static void
number_draw(GtkWidget* widget, GdkRectangle* area, gpointer data)
{
    GdkRectangle dest_area;
    guint x0 = XDISP + widget->allocation.width / FRACTION;
    guint y0 = YDISP + widget->allocation.height / FRACTION;
    NumberObject *no = (NumberObject*)data;

    /* Number color is changed, we need to update the pixmap cache */
    if (no->old_pixel != colors_get_color(COLOR_CHANNUMS)->pixel)
        number_style_set(widget, data);

    dest_area.x = x0;
    dest_area.y = y0;
    dest_area.width = no->pm_width;
    dest_area.height = no->pm_height;

    /* Update number mask if necessary */
    if (widget->allocation.width != no->width || widget->allocation.height != no->height) {
        no->width = widget->allocation.width;
        no->height = widget->allocation.height;
        di_update_mask(widget->window, no->number_pm, no->mask_gc, x0, y0, no->pm_width, no->pm_height,
            widget->allocation.width, widget->allocation.height);
    }

    if (gdk_rectangle_intersect(area, &dest_area, &dest_area))
        di_draw_drawable(custom_drawing_get_drawable(CUSTOM_DRAWING(widget)), no->mask_gc,
            no->drawable, dest_area.x - x0, dest_area.y - y0,
            dest_area.x, dest_area.y, dest_area.width, dest_area.height);
}

void
scope_group_set_mode(ScopeGroup* s, SampleDisplayMode mode)
{
    guint i;

    for (i = 0; i < 32; i++)
        sample_display_set_mode(s->scopes[i], gui_settings.scopes_mode);
}

GtkWidget*
scope_group_new(GdkPixbuf* mutedpic, GdkPixbuf* unmutedpic)
{
    ScopeGroup* s;
    gint i;

    s = g_object_new(scope_group_get_type(), NULL);
    GTK_BOX(s)->spacing = 2;
    GTK_BOX(s)->homogeneous = FALSE;
    s->scopes_on = 0;
    s->update_freq = 40;
    s->gtktimer = -1;
    s->numchan = 2;
    s->on_mask = 0xFFFFFFFF;

    s->table = gtk_table_new(2, 1, TRUE);
    gtk_table_set_col_spacings(GTK_TABLE(s->table), 4);
    gtk_table_set_row_spacings(GTK_TABLE(s->table), 4);
    gtk_widget_show(s->table);
    gtk_box_pack_start(GTK_BOX(s), s->table, TRUE, TRUE, 0);

    for (i = 0; i < 32; i++) {
        GtkWidget *box, *thing;

        box = gtk_vbox_new(FALSE, 0);
        s->scopebuttons[i] = box;
        g_signal_connect(box, "button-press-event",
            G_CALLBACK(scope_group_button_press), GINT_TO_POINTER(i));
        g_object_ref(box);
        gtk_widget_show(box);

        thing = scalable_pic_new(mutedpic, TRUE);
        gtk_widget_add_events(thing, GDK_BUTTON_PRESS_MASK);
        s->number_m[i] = init_number(i + 1, thing);
        custom_drawing_register_drawing_func(CUSTOM_DRAWING(thing), number_draw, s->number_m[i]);
        custom_drawing_register_realize_func(CUSTOM_DRAWING(thing), number_realize, s->number_m[i]);
        custom_drawing_register_style_set_func(CUSTOM_DRAWING(thing), number_style_set, s->number_m[i]);
        colors_add_widget(COLOR_CHANNUMS, thing);
        gtk_box_pack_start(GTK_BOX(box), thing, TRUE, TRUE, 0);
        s->mutedpic[i] = thing;

        thing = scalable_pic_new(unmutedpic, FALSE);
        gtk_widget_add_events(thing, GDK_BUTTON_PRESS_MASK);
        s->number_u[i] = init_number(i + 1, thing);
        custom_drawing_register_drawing_func(CUSTOM_DRAWING(thing), number_draw, s->number_u[i]);
        custom_drawing_register_realize_func(CUSTOM_DRAWING(thing), number_realize, s->number_u[i]);
        custom_drawing_register_style_set_func(CUSTOM_DRAWING(thing), number_style_set, s->number_u[i]);
        colors_add_widget(COLOR_CHANNUMS, thing);
        gtk_box_pack_start(GTK_BOX(box), thing, TRUE, TRUE, 0);
        s->unmutedpic[i] = thing;

        thing = sample_display_new(FALSE, gui_settings.scopes_mode);
        gtk_widget_add_events(thing, GDK_BUTTON_PRESS_MASK);
        s->number_c[i] = init_number(i + 1, thing);
        custom_drawing_register_drawing_func(CUSTOM_DRAWING(thing), number_draw, s->number_c[i]);
        custom_drawing_register_realize_func(CUSTOM_DRAWING(thing), number_realize, s->number_c[i]);
        custom_drawing_register_style_set_func(CUSTOM_DRAWING(thing), number_style_set, s->number_c[i]);
        colors_add_widget(COLOR_CHANNUMS, thing);
        gtk_box_pack_start(GTK_BOX(box), thing, TRUE, TRUE, 0);
        s->scopes[i] = SAMPLE_DISPLAY(thing);
    }

    gtk_table_attach_defaults(GTK_TABLE(s->table), s->scopebuttons[0], 0, 1, 0, 1);
    gtk_table_attach_defaults(GTK_TABLE(s->table), s->scopebuttons[1], 0, 1, 1, 2);

    scope_group_set_update_freq(s, gui_settings.scopes_update_freq);

    return GTK_WIDGET(s);
}

static void
scope_group_realize(GtkWidget *widget)
{
    static gboolean firsttime = TRUE;
    ScopeGroup *s;
    guint i;

    g_return_if_fail(widget != NULL);
    g_return_if_fail(GTK_IS_HBOX(widget));

    GTK_WIDGET_CLASS(scope_group_parent_class)->realize(widget);
    s = SCOPE_GROUP(widget);

    if (firsttime) {
        firsttime = FALSE;
        s->bg_gc = colors_get_gc(COLOR_UNMUTED_BG);

        for (i = 0; i < 32; i++) {
            scalable_pic_set_bg(SCALABLE_PIC(s->unmutedpic[i]), s->bg_gc);
        }
    }
}

static void
scope_group_class_init(ScopeGroupClass *class)
{
    GtkWidgetClass *widget_class = (GtkWidgetClass*)class;

    widget_class->realize = scope_group_realize;
}

static void
scope_group_init (ScopeGroup *self)
{
}
